#include <string>
#include <mpi.h>
#ifndef PADICO
#include <CORBA.h>
#include <Naming.hh>
#include <thread/paco_omni.h>
#else
#include <Padico/CORBA.h>
#include <Padico/COSNaming.h>
#include <thread/paco_marcel.h>
#endif
#include <interfacePaCO.h>
#include <interfacePaCO_Basic_Example_client.h> // interfacePaco+module+interface+_client
#include <com/paco_mpi.h>
#include <comScheduling/paco_direct_comScheduling.h>
#include <stdio.h>
#include <Padico/PadicoTM.h>


#undef  DIST_BASICBC
#define DIST_IDENTITY

#undef USE_LIBGRAPH


#ifdef DIST_BASICBC
#include <BasicBC.h>
#endif
#ifdef DIST_IDENTITY
#include <Identity.h>
#endif
#ifdef USE_LIBGRAPH
#include <comScheduling/paco_libgraph_comScheduling.h>
#endif

static int  myid, numprocs;

static Basic::PaCO_Example * serveur;

#ifndef PADICO
#else
typedef padico_timing_t my_time_t;
#define get_time(a) padico_timing_get_tick(&a)
#define timing_diff_usec(a,b) (padico_timing_diff_usec(&a,&b)/1000)
#endif


void init() {
  // Init PACO
  paco_fabrique_manager* pfm = paco_getFabriqueManager();

  cerr << "[client] registering com"  << endl;
  pfm->register_com("mpi", new paco_mpi_fabrique());
  cerr << "[client] registering thread"  << endl;  
#ifndef PADICO
  pfm->register_thread("mythread", new paco_omni_fabrique());
#else
  pfm->register_thread("mythread", new paco_marcel_fabrique());
#endif

#ifdef DIST_BASICBC
  cerr << "[client] registering BasicBC"  << endl;
  pfm->register_distribution("BasicBC", new BasicBC_fab());
#endif
#ifdef DIST_IDENTITY
  cerr << "[client] registering IDENTITY"  << endl;
  pfm->register_distribution("Identity", new Identity_fab());
#endif

  cerr << "[client] registering direct"  << endl;
  pfm->register_comScheduling("direct", new paco_direct_fabrique());
#ifdef USE_LIBGRAPH
  cerr << "[client] registering libgraph"  << endl;
  pfm->register_comScheduling("libgraph", new paco_libgraph_fabrique());
#endif
}

void init_connection(char comSchedule) {

  cerr << "[client] Etape Init Connection"  << endl;  
  
  // Contexte global
  PaCO_operation * global_ptr = serveur->getContext("global_paco_context");

  cerr << "[client] Setting thread fabrique"  << endl;  
  global_ptr->setLibThread("mythread");

  cerr << "[client] Setting communication fabrique"  << endl;  
  MPI_Comm group = MPI_COMM_WORLD;
  global_ptr->setLibCom("mpi",&group);

  cerr << "[client] Setting comScheduling fabrique"  << endl;  
  switch(comSchedule) {
  case 'd':   global_ptr->setLibComScheduling("direct"); break;

  case 'l':
#ifdef USE_LIBGRAPH
   global_ptr->setLibComScheduling("libgraph"); 
#else
   cerr << "Sorry: not compile with libgraph support\n";
   abort();
#endif
   break;
  default:  cerr << "Unkown comScheduling: d[irect] or l[ibgraph]"<<endl; 
    exit(-1);
  }


  cerr << "[client] Initializing global context"  << endl;    
  global_ptr->setTypeClient(true);

  PaCO::PacoTopology_t topo;
  topo.total = numprocs;
  global_ptr->setClientTopo(topo);
    
  // Contexte local
  cerr << "[client] Initializing local context"  << endl;  
  PaCO_operation * method_ptr = serveur->getContext("work_in");

#ifdef DIST_BASICBC
  method_ptr->setDisLibArg(0,"BasicBC","in");
#endif

#ifdef DIST_IDENTITY
  method_ptr->setDisLibArg(0,"Identity","in");
#endif

  cerr << "[client] Finishing init"  << endl;  
  global_ptr->init(myid, numprocs);

  cerr << "[client] Etape Init Connection done"  << endl;  

}

void latence_mpi(FILE* fout) {


  my_time_t tick_start, tick_end;
  double time_usec;
  
  MPI_Barrier(MPI_COMM_WORLD);
  
  // Chauffer les caches
  for (unsigned i=0; i< 10; i++)
    MPI_Barrier(MPI_COMM_WORLD);

  MPI_Barrier(MPI_COMM_WORLD);

  get_time(tick_start);
  for (unsigned i=0; i< 1000; i++)
    MPI_Barrier(MPI_COMM_WORLD);
  get_time(tick_end);
  time_usec = timing_diff_usec(tick_start, tick_end);
  fprintf(fout, "#Latence MPI (us)\n%f\n",time_usec);
  
  fflush(fout);
}

void latence(FILE* fout) {


  my_time_t tick_start, tick_end;
  double time_usec;
  
  MPI_Barrier(MPI_COMM_WORLD);
  
  // Chauffer les caches
  for (unsigned i=0; i< 10; i++)
    serveur->latence();

  MPI_Barrier(MPI_COMM_WORLD);

  get_time(tick_start);
  for (unsigned i=0; i< 1000; i++)
      serveur->latence();
  get_time(tick_end);
  time_usec = timing_diff_usec(tick_start, tick_end);
  fprintf(fout, "#Latence PaCO++ (us)\n%f\n",time_usec);
  fflush(fout);  
}


void loop(FILE* fout, const unsigned llen, const unsigned nbiter, char comSchedule) {

  cerr << "[client] loop: Creation distribu en bloc des donnes" << endl;

    Basic::Example::lseq donnees;
    donnees.length(llen);

    for (unsigned i=0; i< llen; i++) 
      {
	donnees[i] = i + 1000*myid;
      }
    
    // Contexte local
    PaCO_operation * method_ptr = serveur->getContext("work_in");
    BasicBC * dislib = (BasicBC *) method_ptr->libraries_in[0];

    dislib->setBlocSize(0); // BLOC
    dislib->setEltSize(sizeof(long));
    dislib->setTotalNbElt(llen*numprocs);
    dislib->setNodeRank(myid);

    // Configuration comSchedule
    switch(comSchedule) {
    case 'd':   /* Nothing */ break;
    case 'l': {
      cerr<<"[Client] Init libgraph"<<endl;
      paco_comScheduling_libgraph* libgraph_ptr = 
	static_cast<paco_comScheduling_libgraph*>(method_ptr->my_comScheduling);
      libgraph_ptr->set_beta(96000);
      libgraph_ptr->set_k(numprocs);
    }
      break;
    default:  cerr << "Unkown comScheduling: d[irect] or l[ibgraph]"<<endl;
      exit(-1);
    }

    padico_timing_t tick_start, tick_end;
    double time_usec;
    
    MPI_Barrier(MPI_COMM_WORLD);

    // Do it
    while(1) {
      get_time(tick_start);
      for (unsigned i=0; i< nbiter; i++) 
	{
	  serveur->work_in(donnees);
	}
      get_time(tick_end);
      time_usec = timing_diff_usec(tick_start, tick_end)/nbiter;
      if (myid==0) {
	struct timeval tv;
	gettimeofday(&tv, NULL);

	fprintf(fout, "%ld.%ld: %d B %f ms %f MB/s\n",
		tv.tv_sec, tv.tv_usec,
		llen*numprocs*sizeof(long), 
		time_usec,(sizeof(long)*numprocs*llen)/(time_usec*1000));
	
	fflush(fout);
      }
    }

}


void test(FILE* fout, const unsigned llen, const unsigned nbiter, char comSchedule) {

  //    cerr << "[client] Creation distribu en bloc des donnes" << endl;

    Basic::Example::lseq donnees;
    donnees.length(llen);

    for (unsigned i=0; i< llen; i++) 
      {
	donnees[i] = i + 1000*myid;
      }
    
    // Contexte local
    PaCO_operation * method_ptr = serveur->getContext("work_in");

#ifdef DIST_BASICBC
    BasicBC * dislib = (BasicBC *) method_ptr->libraries_in[0];
    dislib->setBlocSize(0); // BLOC
    dislib->setEltSize(sizeof(long));
    dislib->setTotalNbElt(llen*numprocs);
#endif
#ifdef DIST_IDENTITY
    Identity * dislib = (Identity *) method_ptr->libraries_in[0];
    dislib->setEltSize(sizeof(long));
    dislib->setLocalNbElt(llen);
#endif
 
    dislib->setNodeRank(myid);

   // Configuration comSchedule
    switch(comSchedule) {
    case 'd':   /* Nothing */ break;
    case 'l': {
#ifdef USE_LIBGRAPH
      cerr<<"[Client] Init libgraph"<<endl;
      paco_comScheduling_libgraph* libgraph_ptr = 
	static_cast<paco_comScheduling_libgraph*>(method_ptr->my_comScheduling);
      libgraph_ptr->set_beta(96000);
      libgraph_ptr->set_k(numprocs);
#endif
    }
      break;
    default:  cerr << "Unkown comScheduling: d[irect] or l[ibgraph]"<<endl;
      exit(-1);
    }

    //    method_ptr->setId(1);


    padico_timing_t tick_start, tick_end;
    double time_usec;
    
    MPI_Barrier(MPI_COMM_WORLD);

    for (unsigned i=0; i< 2; i++) {
      serveur->work_in(donnees);
    }

    MPI_Barrier(MPI_COMM_WORLD);

    // Do it
    get_time(tick_start);
    for (unsigned i=0; i< nbiter; i++) 
      {
	serveur->work_in(donnees);
      }
    get_time(tick_end);
    time_usec = timing_diff_usec(tick_start, tick_end)/nbiter;
    if (myid==0) {
      fprintf(fout, "%d %f %f",llen*numprocs*sizeof(long), time_usec,(sizeof(long)*numprocs*llen)/time_usec);
      
      fflush(fout);
    }

#if 0
    if (myid==0) {
      Basic::global_paco_context * global_ptr = static_cast<Basic::global_paco_context *>(serveur->getContext("global_paco_context"));
      double overhead_usec[10];
      // Overhead user -> PaCO
      //overhead_usec[0] = padico_timing_diff_usec(&tick_start, &global_ptr->_ticks[0]);    
      // Overhead comm matrix
      overhead_usec[1] = padico_timing_diff_usec(&global_ptr->_ticks[1],&global_ptr->_ticks[2]);
      // Overhead comm scheduling
    overhead_usec[2] = padico_timing_diff_usec(&global_ptr->_ticks[2],&global_ptr->_ticks[3]);
    // Overhead send data
    overhead_usec[3] = padico_timing_diff_usec(&global_ptr->_ticks[4],&global_ptr->_ticks[5]);
    // Overhead wait return
    overhead_usec[4] = padico_timing_diff_usec(&global_ptr->_ticks[5],&global_ptr->_ticks[7]);
    // Overhead Paco->user
    overhead_usec[5] = padico_timing_diff_usec(&global_ptr->_ticks[7],&tick_end);
    // Overhead total paco
    //    overhead_usec[6] = padico_timing_diff_usec(&global_ptr->_ticks[0],&global_ptr->_ticks[7]);

    // Overhead total paco
    //    overhead_usec[7] = padico_timing_diff_usec(&global_ptr->_ticks[4],&global_ptr->_ticks[6]);
    // Overhead total paco
    //overhead_usec[8] = padico_timing_diff_usec(&global_ptr->_ticks[6],&global_ptr->_ticks[5]);

    fprintf(fout,"    ");
    for(unsigned i=1; i<6; i++) {
      fprintf(fout,"%f ",overhead_usec[i]);
    }
    fprintf(fout,"\n");
    }
#if 0
    for(unsigned i=0; i<6; i++) {
      overhead_usec[i] = padico_timing_diff_usec(&tick_start, 
						 &global_ptr->local_return_layer->global_paco_context_ptr->_ticks[i]);
      fout<<"tick "<<i<<" : "<<overhead_usec[i]<<endl;
    }
#endif    
#endif


}

int main(int argc, char **argv)
{
    // Init
    cerr << "[Client] Initialisation de MPI" << endl;
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    cerr << "[Client] Initialisation de l'ORB" << endl;
#ifndef PADICO
  CORBA::ORB_var orb = CORBA::ORB_init(argc, argv);
#else
  CORBA::ORB_ptr orb = Padico::the_orb;
#endif

    CORBA::Object_var obj = orb->resolve_initial_references("NameService");
    CosNaming::NamingContextExt_var nc =
        CosNaming::NamingContextExt::_narrow(obj);
    cerr << "[client] Recuperation du serveur" << endl;
    obj = nc->resolve_str("ParallelServeur");
    padico_tm_msleep(100*myid); // in sec

    cerr << "[client] Recuperation de l'interface parallele" << endl;
    serveur = (Basic::PaCO_Example *) Basic::PaCO_Example::PaCO_narrow(obj, orb);

    init();

    unsigned len=10000000;
    unsigned nbiter=1000;
    char comSchedule='l';
    char log='c';

    switch(argc) {
    case 5:
      sscanf(argv[4],"%c", &log);
    case 4:
      sscanf(argv[3],"%d", &nbiter);
    case 3:
      sscanf(argv[2],"%d", &len);
    case 2:
      sscanf(argv[1],"%c", &comSchedule);
    case 1:
      break;
    default:
      cerr << "USAGE: Client [local len in long] [#iter]\n";
      exit(1);
    }

    init_connection(comSchedule);

    FILE* fout;
    if (log == 'f' ) {
      char name[200];
      //    sprintf(name,"/auto/cluster/u/cluster/0/user/cperez/paco.perf.%d",numprocs);
      sprintf(name,"/home/cperez/paco.perf.%d",numprocs);
      if ((fout=fopen(name, "w"))==NULL) {
	perror("impossible to open output file\n");
	abort();
      }
    } else {
      fout = stdout;
    }

    latence_mpi(fout);
    PaCO_operation * method_ptr = serveur->getContext("work_in");

    fflush(fout);

#if 1
    // Loop
    if (myid==0)
      fprintf(fout,"# Loop nbiter=%d lenmax=%d schedule=%c - Cache\n",nbiter, len, comSchedule);
    method_ptr->useCommunicationMemoryId(1);
    method_ptr->libraries_in[0]->setComId(1);
    latence(fout);
    method_ptr->clearCommunicationMemoryId(1);
    method_ptr->libraries_in[0]->setComId(0);
    method_ptr->libraries_in[0]->freeComId(1);
    
    method_ptr->useCommunicationMemoryId(1);
    method_ptr->libraries_in[0]->setComId(1);
    loop(fout, 10000000, 1,comSchedule);
#endif

#if 0
    // Sans cache
    method_ptr->noCommunicationMemory();
    method_ptr->libraries_in[0]->setComId(-1); // No memory Id
    if (myid==0)
      fprintf(fout,"# Test nbiter=%d lenmax=%d schedule=%c - no cache\n",nbiter, len, comSchedule);
    latence(fout);
#if 1
    for(len=1; len<=8000000; len*=2)
      test(fout, len, nbiter,comSchedule);
#endif

    fflush(fout);

    // Avec Cache
    if (myid==0)
      fprintf(fout,"# Test nbiter=%d lenmax=%d schedule=%c - Cache\n",nbiter, len, comSchedule);
    method_ptr->useCommunicationMemoryId(1);
    method_ptr->libraries_in[0]->setComId(1);
    latence(fout);
    method_ptr->clearCommunicationMemoryId(1);
    method_ptr->libraries_in[0]->setComId(0);
    method_ptr->libraries_in[0]->freeComId(1);
#if 1
    for(len=1; len<=8000000; len*=2) {
      method_ptr->useCommunicationMemoryId(1);
      method_ptr->libraries_in[0]->setComId(1);
      test(fout, len, nbiter,comSchedule);
      method_ptr->clearCommunicationMemoryId(1);
      method_ptr->libraries_in[0]->setComId(0);
      method_ptr->libraries_in[0]->freeComId(1);
    }
#endif
#endif

    fflush(fout);
    fclose(fout);

    cerr << "[client] all done...exiting..\n";

    // calcul
    MPI_Finalize();
}
